"""
Classes to process ZFS filesystems
"""

from subprocess import check_output, CalledProcessError

__all__ = [ 'snapshots', 'zpool' 'zfs' ]

SNAPSHOT_DATE_FORMAT = '%Y%m%d-%H%M%S'

class ZFSError(Exception):
    pass

def execute(cmd):
    if isinstance(cmd,basestring):
        cmd = cmd.split(' ')

    try:
        output = check_output(cmd)
    except CalledProcessError:
        raise ZFSError('Error running command %s' % ' '.join(str(x) for x in cmd))

    return [x.rstrip() for x in output.rstrip('\n').split('\n')]



"""
ZFS snapshots
"""

from ultimatum.zfs import execute, ZFSError, SNAPSHOT_DATE_FORMAT

class ZFSSnapshot(list):
    def __init__(self, name):
        self.name = name
        try:
            self.volume, self.tag = name.split('@')
        except ValueError:
            raise ZFSError('Invalid snapshot name: %s' % name)

    def __cmp__(self, other):
        if isinstance(other, basestring):
            try:
                tag = self.name.split('@')[1]
                if tag == other:
                    return 0
            except IndexError:
                pass

            return cmp(self.name, other)

        elif not isinstance(other,ZFSSnapshot):
            raise ZFSError("Can't compare ZFSSnapshot to %s object" % type(other))

        if self.volume != other.volume:
            return cmp(self.volume, other.volume)

        try:
            my_date = time.strptime(self.tag, SNAPSHOT_DATE_FORMAT)
            other_date = time.strptime(other.tag, SNAPSHOT_DATE_FORMAT)
            return cmp(my_date, other_date)
        except ValueError:
            pass

        return cmp(self.tag, other.tag)

    def __eq__(self,other):
        return self.__cmp__(other) == 0

    def __ne__(self,other):
        return self.__cmp__(other) != 0

    def __repr__(self):
        return '%s@%s' % (self.volume, self.tag)

    def rename(self,name):
        if name.count('@')==0:
            name = '%s@%s' % (self.volume, name)
        execute('zfs rename %s %s' % (self.name, name))



#!/usr/bin/env python
"""
Implementation of FreeBSD filesystem mount point parsing
"""

import os,re
from subprocess import check_output,CalledProcessError

from systematic.log import Logger,LoggerError
from systematic.filesystems import MountPoint,FileSystemError

PSEUDO_FILESYSTEM = [
    'procfs','devfs',
]

RE_MOUNT = re.compile(r'([^\s]*) on ([^\s]+) \(([^\)]*)\)$')

class MountPoints(dict):
    """
    Mount points for freeBSD filesystems
    """
    def __init__(self):
        dict.__init__(self)
        self.update()

    def update(self):
        """
        Update list of FreeBSD mountpoints based on /sbin/mount output
        """
        try:
            output = check_output(['/sbin/mount'])
        except CalledProcessError:
            raise FileSystemError('Error running /sbin/mount')

        for l in [l for l in output.split('\n') if l.strip()!='']:
            if l[:4] == 'map ':
                continue
            m = RE_MOUNT.match(l)
            if not m:
                continue

            device = m.group(1)
            mountpoint = m.group(2)
            flags = map(lambda x: x.strip(), m.group(3).split(','))
            filesystem = flags[0]
            flags = flags[1:]

            entry = BSDMountPoint(device,mountpoint,filesystem)
            for f in flags:
                entry.flags.set(f,True)
            self[mountpoint] = entry

class BSDMountPoint(MountPoint):
    """
    One BSD mountpoint based on /sbin/mount output line
    Additional attributes:
    """
    def __init__(self,device,mountpoint,filesystem):
        MountPoint.__init__(self,device,mountpoint,filesystem)

    @property
    def is_virtual(self):
        return self.filesystem in PSEUDO_FILESYSTEM

    @property
    def usage(self):
        """
        Check usage percentage for this mountpoint.
        Returns dictionary with usage details.
        """
        if self.filesystem in PSEUDO_FILESYSTEM:
            return {}
        try:
            output = check_output(['df','-k',self.mountpoint])
        except CalledProcessError:
            raise FileSystemError('Error getting usage for %s' % self.mountpoint)

        (header,usage) = output.split('\n',1)
        try:
            usage = ' '.join(usage.split('\n'))
        except ValueError:
            pass

        (fs,size,used,free,percent,mp) = map(lambda x: x.strip(), usage.split())
        percent = percent.rstrip('%')
        return {
            'mountpoint': self.mountpoint,
            'size': long(size),
            'used': long(used),
            'free': long(free),
            'percent': int(percent)
        }



"""
Reading and writing of sysctl variables as dictionaries
"""

from subprocess import check_output,CalledProcessError

class SysCtlError(Exception):
    def _str__(self):
        return self.args[0]

class SysCtlTree(dict):
    def __init__(self,path=None):
        cmd = path is not None and ['sysctl','-e',path] or ['sysctl','-ea']
        try:
            output = check_output(cmd)
        except CalledProcessError:
            raise SysCtlError('Error running command %s' % ' '.join(cmd))
        for l in [l.strip() for l in output.split('\n') if l.strip()!='']:
            try:
                k,v = l.split('=',1)
                self[k] = v
            except ValueError:
                raise SysCtlError('Error parsing line: %s' % l)

if __name__ == '__main__':
    import sys
    for t in sys.argv[1:]:
        t = SysCtlTree(t)
        for k,v in t.items():
            print k,v




import re
import os
import glob

from seine.address import IPv4Address, IPv6Address, parse_address
from seine.whois.arin import ARINReverseIPQuery, WhoisError
from systematic.log import LogEntry, LogFile, LogFileCollection, LogFileError
from systematic.sqlite import SQLiteDatabase, SQLiteError

SSH_LOGINS = [
    re.compile('^Accepted publickey for (?P<user>[^\s]+) from (?P<address>.*) ' +
        'port (?P<port>\d+) (?P<sshversion>.*): (?P<keytype>.*) (?P<key>.*)$'
    ),
]
SSH_ATTEMPTS = [
    re.compile('^Invalid user (?P<user>[^\s]+) from (?P<address>.*)'),
    re.compile('^Failed publickey for (?P<user>[^\s]+) from (?P<address>.*) ' +
        'port (?P<port>\d+) (?P<sshversion>.*) (?P<keytype>.*) (?P<fingerprint>.*)$'
    ),
    re.compile('^error: Received disconnect from (?P<address>[^:+]):.*: Auth fail [preauth]'),
]


SSH_CONNECT = re.compile('^Connection from (?P<address>[^\s]+) port (?P<port>\d+)$')
SSH_ACCEPT_PUBLICKEY = re.compile('^%s$' % ' '.join([
    'Accepted publickey for (?P<username>[^\s]+)',
    'from (?P<address>[^\s]+)',
    'port (?P<port>\d+)',
    '(?P<version>[^:]+):',
    '(?P<keytype>[^\s]+)',
    '(?P<fingerprint>.*)'
]))

SSH_LOGIN = re.compile('User child is on pid (?P<pid>\d+)$')
SSH_DISCONNECT = re.compile('Received disconnect from (?P<address>[^\s]+): .*: disconnected by user')
SSH_INVALID_USER = re.compile('^Invalid user (?P<username>.*) from (?P<address>[^\s]+)$')

SSHD_VIOLATIONS_DATABASE_PATH = '/var/lib/ssh/violations.db'
SQL_TABLES = [
"""CREATE TABLE IF NOT EXISTS registration (
    id              INTEGER PRIMARY KEY,
    version         INT,
    handle          TEXT,
    comment         TEXT,
    registered      DATETIME,
    updated         DATETIME
)""",
"""CREATE TABLE IF NOT EXISTS netblock (
    id              INTEGER PRIMARY KEY,
    registration    INT REFERENCES registration(id) ON DELETE CASCADE,
    description     TEXT,
    network         TEXT,
    start           TEXT,
    end             TEXT
)""",
"""CREATE UNIQUE INDEX IF NOT EXISTS netblock_registration ON netblock(registration, network)""",
"""CREATE TABLE IF NOT EXISTS login (
    id              INTEGER PRIMARY KEY,
    timestamp       DATETIME,
    registration    INT REFERENCES registration(id) ON DELETE CASCADE,
    address         TEXT,
    username        TEXT
)""",
"""CREATE UNIQUE INDEX IF NOT EXISTS attempts ON login(timestamp, address, username)"""
]


class SSHSession(list):
    def __init__(self, sessioncache, entry, pid=None, parent=None, timeout=60):
        self.sessioncache = sessioncache
        self.pid = pid is not None and pid or entry.pid
        self.timeout = timeout
        self.parent = parent
        self.info = {}

        self.append(entry)

    def __repr__(self):
        return self.pid

    def append(self, entry):
        if len(self) == 0:
            self.state = 'init'
        list.append(self, entry)

        m = SSH_CONNECT.match(entry.message)
        if m:
            self.state = 'connect'
            self.info['src_address'] = m.groupdict()['address']
            self.info['src_port'] = m.groupdict()['port']
            return

        m = SSH_ACCEPT_PUBLICKEY.match(entry.message)
        if m:
            self.state = 'accepted_publickey'
            details = m.groupdict()
            for k in ( 'username', 'version', 'keytype', 'fingerprint', ):
                if k in details:
                    self.info[k] = details[k]
                else:
                    self.info[k] = 'MISSING KEY INFO %s' % k
            return

        m = SSH_LOGIN.match(entry.message)
        if m:
            if self.parent is not None:
                self.state = 'user_session'
                self.info['parent_pid'] = self.parent.pid
                self.info.update(self.parent.info.items())
            else:
                self.state = 'login'
                usersession_pid = m.groupdict()['pid']
                if usersession_pid != self.pid:
                    self.sessioncache.append(SSHSession(self.sessioncache, entry, pid=usersession_pid, parent=self))
            return

        m = SSH_DISCONNECT.match(entry.message)
        if m:
            self.state = 'logout'
            if len(self) > 0:
                self.info['session_length'] = (entry.time - self[0].time).total_seconds()
            return

        m = SSH_INVALID_USER.match(entry.message)
        if m:
            self.state = 'invalid_user'
            self.info['username'] = m.groupdict()['username']
            return

        if entry.message == 'fatal: Read from socket failed: Connection reset by peer [preauth]':
            self.state = 'preauth_connection_reset'

        elif 'src_address' in self.info:
            if entry.message == 'Connection closed by %s [preauth]' % self.info['src_address']:
                self.state = 'preauth_no_key'

            if entry.message == 'Received disconnect from %s: 11: Bye Bye [preauth]' % self.info['src_address']:
                if self.state != 'invalid_user':
                    self.state = 'preauth_disconnect'


    def match(self, entry):
        if entry.pid != self.pid:
            return False

        if self.state == 'user_session':
            return True

        if abs((entry.time - self[0].time).total_seconds()) >= self.timeout:
            return False

        return True


class SSHSessionCache(dict):
    def match(self, entry):
        if entry.pid not in self:
            return None

        for session in self[entry.pid]:
            if session.match(entry):
                return session
        return None

    def append(self, session):
        if session.pid not in self:
            self[session.pid] = []
        self[session.pid].append(session)


class AuthLogEntry(LogEntry):
    def __init__(self, *args, **kwargs):
        LogEntry.__init__(self, *args, **kwargs)

        if self.program == 'sshd':
            session = self.logfile.sessioncache.match(self)
            if session is not None:
                session.append(self)
            else:
                session = SSHSession(self.logfile.sessioncache, self)
                self.logfile.sessioncache.append(session)

class AuthLogFile(LogFile):
    lineloader = AuthLogEntry
    def __init__(self, sessioncache, *args, **kwargs):
        LogFile.__init__(self, *args, **kwargs)

        self.sessioncache = sessioncache

        self.register_iterator('failures')
        self.register_iterator('logins')

    def __match_failed__(self, entry):
        for matcher in SSH_ATTEMPTS:
            m = matcher.match(entry.message)
            if m:
                details = m.groupdict()
                entry.update_message_fields(details)
                return True

        return False

    def __match_login__(self, entry):
        for matcher in SSH_LOGINS:
            m = matcher.match(entry.message)
            if m:
                details = m.groupdict()
                if 'port' in details:
                    details['port'] = int(details['port'])
                details['address'] = parse_address(details['address'])
                entry.update_message_fields(details)
                return True

        return False

    def next_failed(self):
        return self.next_iterator_match('failures', callback=self.__match_failed__)

    def next_login(self):
        return self.next_iterator_match('logins', callback=self.__match_login__)

    @property
    def failures(self):
        return iter(self.next_failed, None)

    @property
    def logins(self):
        return iter(self.next_login, None)


class AuthLogCollection(LogFileCollection):
    loader = AuthLogFile
    def __init__(self, *args, **kwargs):
        LogFileCollection.__init__(self, *args, **kwargs)
        self.sessioncache = SSHSessionCache()

class SSHViolationsDatabase(SQLiteDatabase):
    def __init__(self, path=SSHD_VIOLATIONS_DATABASE_PATH):
        SQLiteDatabase.__init__(self, path, SQL_TABLES)

    def lookup_registration_id(self, address):
        try:
            address = IPv4Address(address)
        except ValueError:
            try:
                address = IPv6Address(address)
            except ValueError:
                raise ValueError('ERROR parsing address %s' % address)

        c = self.cursor
        c.execute("""SELECT registration,network FROM netblock""")
        for entry in c.fetchall():
            try:
                network = IPv4Address(entry[1])
            except ValueError:
                try:
                    network = IPv6Address(enrty[1])
                except ValueError:
                    continue

            if type(network) != type(address):
                continue

            if network.hostInNetwork('%s' % address):
                return entry[0]

        return None

    def add_netblock(self, ref):
        c = self.cursor
        c.execute("""SELECT id FROM registration WHERE handle=?""", (ref.handle,))
        r = c.fetchone()
        if r is not None:
            return None

        c.execute("""INSERT INTO registration (version, handle, comment, registered, updated) """ +
            """VALUES (?,?,?,?,?)""",
            (ref.version, ref.handle, ref.comment, ref.registered, ref.updated, )
        )
        self.commit()

        c.execute("""SELECT id FROM registration WHERE handle=?""", (ref.handle,))
        ref_id = int(c.fetchone()[0])

        for netblock in ref:
            if isinstance(netblock.network, IPv4Address):
                c.execute("""INSERT INTO netblock (registration, description, network, start, end) """ +
                    """VALUES (?,?,?,?,?)""",
                    (
                        ref_id,
                        netblock.description,
                        netblock.network.cidr_address,
                        netblock.start.cidr_address,
                        netblock.end.cidr_address,
                    )
                )
            elif isinstance(netblock.network, IPv6Address):
                    (
                        ref_id,
                        netblock.description,
                        '%s' % netblock.network,
                        '%s' % netblock.start,
                        '%s' % netblock.end,
                    )

        self.commit()

        return ref_id

    def add(self, timestamp, address, username, registration):
        c = self.cursor
        c.execute("""SELECT * FROM login WHERE timestamp=? AND address=? AND username=?""",
            ( timestamp, address, username, )
        )
        r = c.fetchone()
        if r is not None:
            return None

        c.execute("""INSERT INTO login (timestamp, registration, address, username) VALUES (?,?,?,?)""",
            ( timestamp, registration, address, username, )
        )
        self.commit()

        c.execute("""SELECT * FROM login WHERE timestamp=? AND address=? AND username=?""",
            ( timestamp, address, username, )
        )
        r = c.fetchone()
        return self.as_dict(c, r)

    def update(self, paths=None):
        matcher = re.compile('^Invalid user (?P<user>[^\s]+) from (?P<address>.*)')
        if not paths:
            auth_paths = glob.glob('/var/log/auth.log*') + glob.glob('/var/log/messages*')

        sessioncache = SSHSessionCache()
        for path in paths:
            log = AuthLogFile(sessioncache, path)
            for entry in log.failures:
                details = {
                    'timestamp': entry.time,
                    'address': entry.message_fields['address'],
                    'username': entry.message_fields['user'],
                }

                ref = self.lookup_registration_id(details['address'])
                if ref is not None:
                    details['registration'] = ref

                elif isinstance(details['address'], IPv4Address):
                    self.log.debug('ARIN LOOKUP %s' % details['address'])
                    ref = ARINReverseIPQuery(details['address'])
                    details['registration'] = self.add_netblock(ref)

                else:
                    details['registration'] = None

                self.add(**details)

    def map_netblocks(self, values):
        c = self.cursor
        c.execute("""SELECT registration,network FROM netblock ORDER BY registration""")

        registration_netblock_map = {}
        for nb in c.fetchall():
            if nb[0] not in registration_netblock_map:
                registration_netblock_map[nb[0]] = []
            try:
                address = IPv4Address(nb[1])
            except ValueError:
                try:
                    address = IPv6Address(nb[1])
                except ValueError:
                    continue

            if address not in registration_netblock_map[nb[0]]:
                registration_netblock_map[nb[0]].append(address)

        for reg in registration_netblock_map:
            registration_netblock_map[reg].sort()

        for value in values:
            value['netblocks'] = []
            if value['registration'] in registration_netblock_map:
                for nb in registration_netblock_map[value['registration']]:
                    if nb.hostInNetwork(value['address']):
                        value['netblocks'].append(nb)

        return values

    def source_address_counts(self):
        c = self.cursor
        c.execute("""SELECT COUNT(DISTINCT timestamp) AS count, registration, address """ +
            """FROM login GROUP BY address ORDER BY -count"""
        )
        return self.map_netblocks([self.as_dict(c,r) for r in c.fetchall()])

    def login_attempts(self, start=None):
        c = self.cursor

        if start is not None:
            c.execute("""SELECT * FROM login WHERE timestamp >= Datetime(?) """ +
                """ORDER BY timestamp""",
                (start,)
            )
        else:
            c.execute("""SELECT * FROM login ORDER BY timestamp""")

        return self.map_netblocks([self.as_dict(c, r) for r in c.fetchall()])




"""
ZFS API for python
"""

import logging
from datetime import datetime, timedelta
from subprocess import Popen, PIPE

from ultimatum.zfs import execute, ZFSError, SNAPSHOT_DATE_FORMAT
from ultimatum.zfs.snapshots import ZFSSnapshot

ZFS_BOOLEAN_PROPERTIES = (
    'atime',
    'checksum',
    'canmount',
    'dedup',
    'devices',
    'exec',
    'jailed',
    'mounted',
    'nbmand',
    'readonly',
    'setuid',
    'sharenfs',
    'sharesmb',
    'type',
    'utf8only',
    'vscan',
    'xattr'
)
ZFS_READONLY_PROPERTIES = (
    'available',
    'creation',
    'refcompressratio',
    'referenced',
    'type',
    'used',
    'usedbychildren',
    'usedbydataset',
    'usedbyrefreservation',
    'usedbysnapshots',
    'written',
)
ZFS_STRING_PROPERTIES =  (
    'aclinherit',
    'aclmode',
    'casesensitivity',
    'compressratio',
    'copies',
    'logbias',
    'mlslabel',
    'mountpoint',
    'normalization',
    'primarycache',
    'quota',
    'recordsize',
    'refquota',
    'refreservation',
    'reservation',
    'secondarycache',
    'snapdir',
    'sync',
    'version',
)

# TODO - process these flags properly as in zpool.py
"""
aclinherit restricted
aclmode discard
casesensitivity sensitive
compressratio 1.00x
copies 1
logbias latency
mlslabel -
mountpoint none
normalization none
primarycache all
quota none
recordsize 128K
refquota none
refreservation none
reservation none
secondarycache all
snapdir hidden
sync standard
version 4
"""

ZFS_PROPERTY_VALIDATORS = {

}

ZFS_PROPERTIES = ZFS_BOOLEAN_PROPERTIES + ZFS_READONLY_PROPERTIES + ZFS_STRING_PROPERTIES

class ZFS(object):
    """ZFS filesystem object

    Wrapper to manipulate a ZFS filesystem object

    """
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return 'zfs %s' % self.name

    @property
    def snapshots(self):
        """List of snapshots

        Returns list of existing snapshots as ZFSSnapshot objects

        """
        snapshots = []
        for name in execute('zfs list -Hrt snapshot -o name %s' % self.name):
            if name == '':
                continue

            snapshot = ZFSSnapshot(name)
            if snapshot.volume == self.name:
                snapshots.append(snapshot)

        return snapshots

    def get_property(self, key):
        """Return property value

        Return value for a ZFS property, validated and formatted

        """
        if key not in ZFS_PROPERTIES:
            raise ZFSError('Invalid property name')

        value = None
        cmd = 'zfs get -H %s %s' % (key, self.name)
        for line in execute(cmd):
            try:
                volume, field, value, flags = line.split(None, 3)
                if volume == self.name and field == key:
                    break
            except ValueError:
                continue

            value = None

        if value in (None, 'none',):
            return None

        if value == '-' and key in ZFS_OPTIONAL_PROPERTIES:
            return None

        if key in ZFS_BOOLEAN_PROPERTIES:
            return value == 'on'

        return value

    def set_property(self, key, value):
        """Set property value

        Set value for a ZFS property, validated and formatted

        """
        skip_quotes="True"
        if key not in ZFS_PROPERTIES:
            raise ZFSError('Invalid property name')

        if key in ZFS_READONLY_PROPERTIES:
            raise ZFSError('Readonly property: %s' % key)

        if key in ZFS_PROPERTY_VALIDATORS:
            if not ZFS_PROPERTY_VALIDATORS[key](value):
                raise ZFSError('Unknown value for property %s: %s' % (key, value))

        if key in ZFS_BOOLEAN_PROPERTIES:
            value = value and 'on' or 'off'
            skip_quotes = True

        elif value == None:
            value = 'none'

        if skip_quotes:
            execute(['zfs', 'set', '%s=%s' % (key, value), self.name])
        else:
            execute(['zfs', 'set', '%s="%s"' % (key, value), self.name])

    def create_snapshot(self, tag=None):
        """Create named snapshots

        Creates a new snapshot with provided name

        Raises ZFSError if named snapshot exists
        """
        if tag is None:
            tag = datetime.now().strftime(SNAPSHOT_DATE_FORMAT)

        name = '%s@%s' % (self.name, tag)
        if name in self.snapshots:
            raise ZFSError('Snapshot already exists: %s' % name)

        execute('zfs snapshot %s' % name)
        return tag

    def remove_snapshot(self, value):
        """Remove existing snapshot

        Remove a existing snapshot matching provided name

        Raises ZFSError if provided snapshot did not exist
        """
        if isinstance(value, basestring):
            name = '%s@%s' % (self.name, value)
            if name not in self.snapshots:
                raise ZFSError('No such snapshot: %s' % name)
        elif isinstance(value, ZFSSnapshot):
            name = value.name

        execute('zfs destroy %s' % name)

    def get_snapshot(self, name):
        """Lookup snapshot by name

        Return snapshot by name, or None if snapshot was not defined

        """
        for snapshot in self.snapshots:
            if snapshot.name == name:
                return snapshot
        return None

    def filter_snapshots(self, start, stop, date_format=SNAPSHOT_DATE_FORMAT):
        """Filter snapshots by date

        Retrun list of snapshots matching start and stop date ranges in name

        """
        try:
            if not isinstance(start, datetime):
                start = datetime.strptime(start, date_format)
            if not isinstance(stop, datetime):
                stop = datetime.strptime(stop, date_format)

        except ValueError, emsg:
            raise ZFSError('Filter dates do not match default date format: %s' % SNAPSHOT_DATE_FORMAT)

        if start > stop:
            raise ZFSError('Invalid date range: start is after stop')

        snapshots = []
        for snapshot in self.snapshots:
            try:
                ss_date = datetime.strptime(snapshot.tag, date_format)
            except ValueError:
                # Ignore snapshot not matching date formatting
                continue

            if ss_date >= start and ss_date <= stop:
                snapshots.append(snapshot)

        return snapshots

    def clone_to_pool(self, pool, tag, force=False):
        """Clone filesystem to other pool

        Clone this filesystem to target zpool
        """

        if self.snapshots:
            latest = '%s' % self.snapshots[-1]
            self.create_snapshot(tag)
            src_args = ['zfs', 'send', '-i', latest, '%s@%s' % (self.name, tag)]
        else:
            self.create_snapshot(tag)
            src_args = ['zfs', 'send', '%s@%s' % (self.name, tag)]

        if force:
            dst_args = ['zfs', 'receive', '-Fd', pool.name]
        else:
            dst_args = ['zfs', 'receive', '-d', pool.name]

        src = Popen(src_args, stdout=PIPE)
        dst = Popen(dst_args, stdin= src.stdout, stdout=PIPE)
        output = dst.communicate()[0]



#!/usr/bin/env python
"""
Scripts and system tool wrappers for FreeBSD (and other *BSDs)

This module is split from module to platform dependent tool
"""

import glob

from setuptools import setup, find_packages

VERSION='4.0.7'

setup(
    name = 'ultimatum',
    keywords = 'System Management Utility FreeBSD Scripts',
    description = 'Sysadmin utility modules and scripts for FreeBSD',
    author = 'Ilkka Tuohela',
    author_email = 'hile@iki.fi',
    version = VERSION,
    url = 'https://github.com/hile/ultimatum',
    license = 'PSF',
    zip_safe = False,
    packages = find_packages(),
    scripts = glob.glob('bin/*'),
    install_requires = (
        'seine>=3.0.2',
        'systematic>=4.2.6',
    ),
)




"""
Zpool API
"""

import logging
from datetime import datetime, timedelta

from ultimatum.zfs import execute, ZFSError, SNAPSHOT_DATE_FORMAT
from ultimatum.zfs.zfs import execute, ZFS

ZPOOL_READONLY_PROPERTIES = (
    'allocated',
    'capacity',
    'dedupratio',
    'expandsize',
    'free',
    'guid',
    'health',
    'size',
    'readonly',
)
ZPOOL_BOOLEAN_PROPERTIES = (
    'autoexpand',
    'autoreplace',
    'delegation',
    'listsnapshots',
    'readonly',
)
ZPOOL_STRING_PROPERTIES =  (
    'altroot',
    'bootfs',
    'cachefile',
    'comment',
    'dedupditto',
    'failmode',
    'version',
)
ZPOOL_OPTIONAL_PROPERTIES = {
    'altroot',
    'bootfs',
    'cachefile',
    'comment',
    'expandsize',
}
ZPOOL_HEALTH_STATES = (
    'DEGRADED', 'FAULTED', 'OFFLINE', 'ONLINE', 'REMOVED', 'UNAVAIL'
)
ZPOOL_PROPERTY_VALIDATORS = {
    'altroot':  lambda x: isinstance(x, basestring),
    'bootfs': lambda x: isinstance(x, basestring) and x.count('/')>0,
    'cachefile': lambda x: isinstance(x, int),
    'dedupditto': lambda x: isinstance(x, int),
    'failmode': lambda x: x in ( 'wait', 'continue', 'panic' ),
    'dedupditto': lambda x: isinstance(x, int),
    'version': lambda x: isinstance(x, int),
}
ZPOOL_PROPERTIES = ZPOOL_READONLY_PROPERTIES + ZPOOL_BOOLEAN_PROPERTIES + ZPOOL_STRING_PROPERTIES

logger = logging.getLogger(__name__)

def poolnames():
    """Pool name list

    Return list of ZPool names

    """
    names = []
    for line in execute('zpool list -H'):
        names.append(line.split()[0])
    return names

class ZPool(object):
    """ZPool object

    Abstraction of zpool commands for python

    """
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return 'zpool %s' % self.name

    def get_property(self, property):
        """Return property value

        Return value of a zpool entry, validated and formatted

        """
        if property not in ZPOOL_PROPERTIES:
            raise ZFSError('Invalid property name')

        value = ' '.join(execute('zpool list -H -o %s %s' % (property, self.name)))

        if value == '-' and property in ZPOOL_OPTIONAL_PROPERTIES:
            return None

        if property in ZPOOL_BOOLEAN_PROPERTIES:
            return value == 'on'

        if property == 'health' and value not in ZPOOL_HEALTH_STATES:
            raise ZFSError('Unknown health state for pool %s: %s' % (self.name, value))

        return value

    def set_property(self, property, value):
        """Set property value

        Set value of a zpool entry, validated and formatted

        """
        if property not in ZPOOL_PROPERTIES:
            raise ZFSError('Invalid property name')

        if property in ZPOOL_READONLY_PROPERTIES:
            raise ZFSError('Readonly property: %s' % property)

        if property in ZPOOL_PROPERTY_VALIDATORS:
            if not ZPOOL_PROPERTY_VALIDATORS[property](value):
                raise ZFSError('Unknown value for property %s: %s' % (property, value))

        if property in ZPOOL_BOOLEAN_PROPERTIES:
            value = value and 'on' or 'off'

        elif value == None:
            value = 'none'

        execute(['zpool', 'set', '%s="%s"' % (property, value), self.name])

    @property
    def is_available(self):
        """Check if pool is is_available

        Boolean property to see if this pool is available

        """
        if self.name not in poolnames():
            return False

        return True

    @property
    def filesystems(self):
        """List of filesystems

        Return list of ZFS objects for filesystems in this pool

        """
        return [ZFS(fs) for fs in execute('zfs list -Hr -o name %s' % self.name) if fs!='']

    def import_pool(self):
        """Import zpool

        Try to import this zpool

        """
        execute('zpool import %s' % self.name)

    def export_pool(self):
        """Export zpool

        Try to export this zpool

        """
        execute('zpool export %s' % self.name)

    def create_snapshots(self, tag=None):
        """Create snapshots

        Attempt to create named snapshots for all filesystems in this pool

        If tag is not provided, we use a timestamp with SNAPSHOT_DATE_FORMAT

        """
        if tag is None:
            tag = datetime.now().strftime(SNAPSHOT_DATE_FORMAT)
        for fs in self.filesystems:
            fs.create_snapshot(tag)

    def filter_snapshots(self, start, stop, date_format=SNAPSHOT_DATE_FORMAT):
        """Filter snapshots by name

        Filter snapshots of pool's filesystems by date range

        """
        try:
            if not isinstance(start, datetime):
                start = datetime.strptime(start, date_format)
            if not isinstance(stop, datetime):
                stop = datetime.strptime(stop, date_format)

        except ValueError, emsg:
            raise ZFSError('Filter dates do not match default date format: %s' % SNAPSHOT_DATE_FORMAT)

        if start > stop:
            raise ZFSError('Invalid range: start date is after stop date')

        snapshots = []
        for fs in self.filesystems:
            snapshots.extend(fs.filter_snapshots(start, stop, date_format))

        return snapshots
